class CaptchaGuess:
    def __init__(self, captcha_token, captchaObfuscationToken, action, guess):
        self.captcha_token = captcha_token
        self.captchaObfuscationToken = captchaObfuscationToken
        self.action = action
        self.guess = guess



from bs4 import BeautifulSoup
from pyotp import TOTP
from datetime import datetime, timedelta
import json
from . import BaseClient


class ReauthException(Exception):
    pass


class Client(BaseClient):
    """
    A low-level client representing Support:

    .. code-block:: python

        import coto

        session = coto.Session()
        client = session.client('support')

    These are the available methods:

    * :py:meth:`get_support_level`
    * :py:meth:`update_support_level`
    """
    def __init__(self, session):
        super().__init__(session)
        self.__xsrf_token = None

    def _url(self, api):
        return "https://console.aws.amazon.com/support/plans/service/{0}?state=hashArgs%23".format(api)

    def _xsrf_token(self):
        if self.__xsrf_token is None:
            self._get_xsrf_token()

        return self.__xsrf_token

    def _get_xsrf_token(self):
        r = self.session()._get(
            'https://console.aws.amazon.com/support/plans/home?region=eu-central-1&state=hashArgs%23'
        )
        
        if r.status_code != 200:
            raise Exception("failed get support xsrf token")

        for cookie in r.cookies:
            if cookie.name == 'XSRF-TOKEN':
                self.__xsrf_token = cookie.value
                break

    def _get(self, api):
        r = self.session()._get(
            self._url(api), headers={'X-XSRF-TOKEN': self._xsrf_token()})

        if 'X-CSRF-TOKEN' in r.headers:
            self.__xsrf_token = r.headers['X-CSRF-Token']

        if r.status_code != 200:
            raise Exception("failed get {0}".format(api))

        return json.loads(r.text)

    def _post(self, api, data=None):
        r = self.session()._post(
            self._url(api),
            headers={
                'Content-Type': 'application/json',
                'X-XSRF-TOKEN': self._xsrf_token(),
            },
            data=json.dumps(data) if data is not None else None,
        )

        if 'X-CSRF-Token' in r.headers:
            self.__xsrf_token = r.headers['X-CSRF-TOKEN']

        if r.status_code != 200:
            raise Exception("failed post {0}".format(api))

        return json.loads(r.text)

    def get_support_level(self):
        """
        Lists the current support contract level for the account.

        Request Syntax:
            .. code-block:: python

                response = client.get_support_level()

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'supportLevel': str,
                    'canChange': bool
                }
        """
        r = self._post('describeSupportLevelSummary', { "lang": "en" })
        return {
            'supportLevel': r['response']['supportLevel'],
            'canChange': r['response']['canChange']
        }

    def update_support_level(self, support_level):
        """
        Change the support contract level for the account.

        Request Syntax:
            .. code-block:: python

                response = client.set_tax_registration(
                    support_level=str,
                )

        Args:
            support_level (str): Desired support contract level.

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'supportLevel': str
                }
        """
        r = self._post('updateSupportLevel', { "supportLevel": support_level })
        return {
            'supportLevel': r['response']['supportLevel']
        }



import uuid
import sys
import base64
import urllib.request

def url_to_base64(url):
    return base64.b64encode(urllib.request.urlopen(url).read()).decode()

def show_image(b64):
    sys.stdout.write("\033]1337;File=")
    sys.stdout.write("name=captcha;inline=1:")
    sys.stdout.write(b64)
    sys.stdout.write("\a\n")

class iTermSolver:
    def __init__(self):
        self.jobs = {}
    
    def solve(self, base64=None, url=None):
        job_id = uuid.uuid4()

        if base64:
            b64 = base64
        elif url:
            b64 = url_to_base64(url)
        else:
            raise Exception("pass `url` or `base64`")

        self.jobs[job_id] = None
        show_image(b64)
        self.jobs[job_id] = input("Guess: ")

        return job_id
        
    def result(self, job_id):
        job_id = uuid.UUID(str(job_id))
        if job_id in self.jobs:
            return self.jobs[job_id]
        else:
            return None

    def incorrect(self, job_id):
        print("You guessed wrong!")



from bs4 import BeautifulSoup
from pyotp import TOTP
from urllib import parse
import json
from . import BaseClient
from .signin_amazon import ap_url
import time


class Client(BaseClient):
    REQUIRES_AUTHENTICATION = False
    __reset_page = None
    _REDIRECT_URL = "https://console.aws.amazon.com/console/home?state=hashArgs%23&isauthcode=true"

    def __init__(self, session):
        super().__init__(session)
        self.__csrf_token = None

    def _csrf_token(self):
        if self.__csrf_token == None:
            self._get_tokens()

        return self.__csrf_token

    def _get_tokens(self):
        r = self.session()._get(
            'https://signin.aws.amazon.com/resetpassword'
        )

        if r.status_code != 200:
            raise Exception("failed get tokens")

        soup = BeautifulSoup(r.text, 'html.parser')
        meta = {
            m['name']: m['content']
            for m in soup.find_all('meta') if 'name' in m.attrs
        }

        if not 'csrf_token' in meta:
            raise Exception("failed get csrf_token")
        self.__csrf_token = meta['csrf_token']

    def _action(self, action, data=None, api="signin"):
        """
        Execute an action on the signin API.

        Args:
            action: Action to execute.
            data: Arguments for the action.

        Returns:
            dict: Action response.
        """
        if not data:
            data = {}

        data['action'] = action
        # data['redirect_uri'] = self._REDIRECT_URL
        data['csrf'] = self._csrf_token()

        r = self.session()._post(
            "https://signin.aws.amazon.com/{0}".format(api),
            data=data,
        )

        if r.status_code != 200:
            print(r.text)
            raise Exception("failed action {0}".format(action))

        out = json.loads(r.text)
        if out['state'].lower() != 'success':
            if 'Message' in out['properties']:
                raise Exception("failed action {0}: {1}".format(action, out['properties']['Message']))
            else:
                raise Exception("failed action {0}".format(action))

        return out['properties']

    def reset_password(self, reset_token_url, password):
        """
        Performs a password reset.
        """
        query = parse.parse_qs(parse.urlparse(reset_token_url).query)
        return self._action('resetPasswordSubmitForm', {
            'token': query['token'][0],
            'key': query['key'][0],
            'newpassword': password,
            'confirmpassword': password,
        }, api='resetpassword')

    def request_otp_forgot_password(self, email):
        """
        Request an OTP to be sent to the email.
        """
        response = self.session()._get(ap_url(email, 'forgotpassword'))
        soup = BeautifulSoup(response.text, 'html.parser')

        error = soup.find(id="message_error")
        if error:
            message = error.get_text()
            # Enter the characters as they are given in the challenge.
            raise Exception(message)

        form = soup.find('form', id="ap_fpp_1a_form")

        data = {'metadata1': self.session()._metadata1_generator.generate()}

        for field in form.find_all('input'):
            name = field.get('name')
            if not name:
                continue
            value = field.get('value')
            data[name] = value
                
        data['email'] = email
        captcha_page = self.session()._post(
            form.get('action'),
            data=data
        )

        captcha_page_soup = BeautifulSoup(captcha_page.text, 'html.parser')
        div = captcha_page_soup.find_all('div', class_='cvf-captcha-img')
        solver = self.session()._captcha_solver
        guess_uuid = solver.solve(url=div[0].img['src'])

        while True:
            guess = solver.result(guess_uuid)

            if guess is None:
                time.sleep(5)
            else:
                break

        error = captcha_page_soup.find(id="message_error")
        if error:
            message = error.get_text()
            # Enter the characters as they are given in the challenge.
            raise Exception(message)

        form = captcha_page_soup.find('form', class_='cvf-widget-form-captcha')

        data = {'metadata1': self.session()._metadata1_generator.generate()}

        for field in form.find_all('input'):
            name = field.get('name')
            if not name:
                continue
            value = field.get('value')
            data[name] = value
                
        data['cvf_captcha_input'] = guess
        verify = self.session()._post(
            "https://www.amazon.com/ap/cvf/verify",
            data=data
        )
        soup = BeautifulSoup(verify.text, 'html.parser')
        if soup.find_all(class_='cvf-widget-alert-id-cvf-captcha-error'):
            try:
                solver.incorrect(guess_uuid)
            except Exception as e:
                print (f"ERROR Reporting {e}")
            return self.request_otp_forgot_password(email)

        self.__reset_page = self.session()._get(
            verify.url
        )
        return self.__reset_page

    def retrieve_otp_from_email(self, content):
        """
        Parses the AWS Email to retrieve the OTP.
        """
        soup = BeautifulSoup(content, 'html.parser')
        otp = soup.find(id="verificationMsg").find(class_='otp').contents[0]
        return otp

    def reset_password_coupled(self, password, otp, request=None):
        """
        Performs a password reset in the Coupled Account.
        """
        if not request and not self.__reset_page:
            raise ValueError('Missing request information')
        if not request:
            request = self.__reset_page

        soup = BeautifulSoup(request.text, 'html.parser')
        form = soup.find(id="verification-code-form")

        data = {'metadata1': self.session()._metadata1_generator.generate()}
        for field in form.find_all('input'):
            name = field.get('name')
            if not name:
                continue
            value = field.get('value')
            data[name] = value
        
        data['code'] = otp

        verify = self.session()._post(
            "https://www.amazon.com/ap/cvf/verify",
            data=data
        )

        reset_password = BeautifulSoup(verify.text, 'html.parser')

        form = reset_password.find('form', id="ap_fpp_1d_form")
        data = {}
        for field in form.find_all('input'):
            name = field.get('name')
            if not name:
                continue
            value = field.get('value')
            data[name] = value
        data['password'] = password
        data['passwordCheck'] = password

        submit_password = self.session()._post(
            form.get("action"),
            data=data
        )
        soup_submit_password = BeautifulSoup(submit_password.text, 'html.parser')

        if soup_submit_password.find_all('div', id="message_success"):
            return True
        else:
            return False



import json
from . import BaseClient


class Client(BaseClient):
    """
    A low-level client representing Biling:

    .. code-block:: python

        import coto

        session = coto.Session()
        client = session.client('billing')

    These are the available methods:

    * Account:
        * :py:meth:`account_status`
        * :py:meth:`close_account`

    * Alternate Contacts:
        * :py:meth:`list_alternate_contacts`
        * :py:meth:`set_alternate_contacts`

    * Tax Registrations:
        * :py:meth:`list_tax_registrations`
        * :py:meth:`set_tax_registration`
        * :py:meth:`delete_tax_registration`
    """

    def __init__(self, session):
        super().__init__(session)
        self.__xsrf_token = None

    def _xsrf_token(self):
        if self.__xsrf_token is None:
            self.__xsrf_token = self._get_xsrf_token()

        return self.__xsrf_token

    def _get_xsrf_token(self):
        r = self.session()._get(
            'https://console.aws.amazon.com/billing/home?region=eu-central-1&state=hashArgs%23'
        )

        if r.status_code != 200:
            raise Exception("failed get billing xsrf token")

        return r.headers['x-awsbc-xsrf-token']

    def _get(self, api):
        r = self.session()._get(
            "https://console.aws.amazon.com/billing/rest/v1.0/{0}?state=hashArgs%23".
            format(api),
            headers={'x-awsbc-xsrf-token': self._xsrf_token()})

        if r.status_code != 200:
            raise Exception("failed get {0}".format(api))

        return r

    def _put(self, api, data=None):
        if data is None:
            r = self.session()._put(
                "https://console.aws.amazon.com/billing/rest/v1.0/{0}?state=hashArgs%23".
                format(api),
                headers={
                    'x-awsbc-xsrf-token': self._xsrf_token(),
                    'Content-Type': 'application/json',
                })
        else:
            r = self.session()._put(
                "https://console.aws.amazon.com/billing/rest/v1.0/{0}?state=hashArgs%23".
                format(api),
                headers={
                    'x-awsbc-xsrf-token': self._xsrf_token(),
                    'Content-Type': 'application/json',
                },
                data=json.dumps(data))

        if r.status_code != 200:
            raise Exception("failed put {}: {}".format(api, r.text))

        return r

    # billing api

    def list_alternate_contacts(self):
        """
        Lists the alternate contacts set for the account. In order to keep the
        right people in the loop, you can add an alternate contact for Billing,
        Operations, and Security communications.

        Request Syntax:
            .. code-block:: python

                response = client.list_alternate_contacts()

        Returns:
            dict: Response Syntax

            .. code-block:: python

                [
                    {
                        'contactId': int,
                        'contactType': 'billing' | 'operations' | 'security',
                        'email': str,
                        'name': str,
                        'phoneNumber': str,
                        'title': str
                    },
                ]
        """
        r = self._get('additionalcontacts')
        return json.loads(r.text)

    def set_alternate_contacts(self, AlternateContacts):
        """
        Sets the alternate contacts set for the account. In order to keep the
        right people in the loop, you can add an alternate contact for Billing,
        Operations, and Security communications.

        Please note that, the primary account holder will continue to receive
        all email communications.

        Contact Types:
            ``billing``:
                The alternate Billing contact will receive billing-related
                notifications, such as invoice availability notifications.
            ``operations``:
                The alternate Operations contact will receive
                operations-related notifications.
            ``security``:
                The alternate Security contact will receive
                security-related notifications. For additional AWS
                security-related notifications, please access the Security
                Bulletins RSS Feed.

        Request Syntax:
            .. code-block:: python

                response = client.set_alternate_contacts(
                    AlternateContacts=[
                        {
                            'contactType': 'billing',
                            'email': str,
                            'name': str,
                            'phoneNumber': str,
                            'title': str
                        },
                        {
                            'contactType': 'operations',
                            'email': str,
                            'name': str,
                            'phoneNumber': str,
                            'title': str
                        },
                        {
                            'contactType': 'security',
                            'email': str,
                            'name': str,
                            'phoneNumber': str,
                            'title': str
                        },
                    ]
                )

        Args:
            AlternateContacts (list): List of alternate contacts.
        """
        self._put('additionalcontacts', AlternateContacts)

    def list_tax_registrations(self):
        """
        Lists the tax registrations set for the account.
        Set your tax information so that your 1099K or W-88EN is generated
        appropriately. Setting this information up also allows you to sell more
        than 200 transactions or $20,000 in Reserved Instances.

        Status:
            ``Verified``:
                Verified
            ``Pending``:
                Pending
            ``Deleted``:
                Deleted

        Request Syntax:
            .. code-block:: python

                response = client.list_tax_registrations()

        Returns:
            dict: Response Syntax

            .. code-block:: python

                [
                    {
                        'address': {
                            'addressLine1': str,
                            'addressLine2': str,
                            'city': str,
                            'countryCode': str,
                            'postalCode': str,
                            'state': str,
                        },
                        'authority': {
                            'country': str,
                            'state': str
                        },
                        'currentStatus': 'Verified' | 'Pending',
                        'legalName': str,
                        'localTaxRegistration': bool,
                        'registrationId': str
                    },
                ]
        """
        r = self._get('taxexemption/eu/vat/information')
        return json.loads(r.text)['taxRegistrationList']

    def set_tax_registration(self, TaxRegistration):
        """
        Set the tax registrations for the account.
        Set your tax information so that your 1099K or W-88EN is generated
        appropriately. Setting this information up also allows you to sell more
        than 200 transactions or $20,000 in Reserved Instances.

        Request Syntax:
            .. code-block:: python

                response = client.set_tax_registration(
                    TaxRegistration={
                        'address': {
                            'addressLine1': str,
                            'addressLine2': str,
                            'city': str,
                            'countryCode': str,
                            'postalCode': str,
                            'state': str,
                        },
                        'authority': {
                            'country': str,
                            'state': str
                        },
                        'legalName': str,
                        'localTaxRegistration': bool,
                        'registrationId': str,
                    }
                )

        Args:
            TaxRegistration (dict): Desired tax registration.
        """
        self._put('taxexemption/eu/vat/information', TaxRegistration)

    def delete_tax_registration(self, TaxRegistration):
        """
        Delete the given tax registrations from the account.

        Request Syntax:
            .. code-block:: python

                response = client.delete_tax_registration(
                    TaxRegistration={
                        'address': {
                            'addressLine1': str,
                            'addressLine2': str,
                            'city': str,
                            'countryCode': str,
                            'postalCode': str,
                            'state': str,
                        },
                        'authority': {
                            'country': str,
                            'state': str
                        },
                        'legalName': str,
                        'localTaxRegistration': bool,
                        'registrationId': str,
                    }
                )

        Args:
            TaxRegistration (dict): Tax registration to delete.
        """
        TaxRegistration['currentStatus'] = 'Deleted'
        return self.set_tax_registration(TaxRegistration)

    def account_status(self):
        """
        Obtain the status of the account.

        Status:
            ``ACTIVE``:
                Active
            ``SUSPENDED``:
                Suspended, will be deleted within 90 days

        Request Syntax:
            .. code-block:: python

                response = client.account_status()

        Returns:
            string: status
        """
        r = self._get('account/status')
        return json.loads(r.text)

    def close_account(self):
        """
        Close the account. Returns True iff successful, otherwise throws
        an exception.

        Request Syntax:
            .. code-block:: python

                client.close_account()
        
        Returns:
            boolean: success
        """
        self._put('account')
        return True



from bs4 import BeautifulSoup
from pyotp import TOTP
from datetime import datetime, timedelta
import json
from . import BaseClient


class ReauthException(Exception):
    pass


class Client(BaseClient):
    """
    A low-level client representing Account:

    .. code-block:: python

        import coto

        session = coto.Session()
        client = session.client('account')

    These are the available methods:

    * :py:meth:`get_account_info`
    * :py:meth:`update_account_name`
    * :py:meth:`update_account_email`
    * :py:meth:`update_account_password`
    """
    _REDIRECT_URL = "https://console.aws.amazon.com/console/home?state=hashArgs%23&isauthcode=true"

    def __init__(self, session):
        super().__init__(session)
        self.__csrf_token = None

    def _csrf_token(self):
        if self.__csrf_token == None:
            self._get_tokens()

        return self.__csrf_token

    def _get_tokens(self):
        r = self.session()._get(
            'https://signin.aws.amazon.com/updateaccount?redirect_uri=https%3A%2F%2Fconsole.aws.amazon.com%2Fbilling%2Fhome%23%2Faccount'
        )

        if r.status_code != 200:
            raise Exception("failed get tokens")

        soup = BeautifulSoup(r.text, 'html.parser')
        meta = {
            m['name']: m['content']
            for m in soup.find_all('meta') if 'name' in m.attrs
        }
        self.__csrf_token = meta['csrf_token']

    def _action(self, action, data=None):
        """
        Execute an action on the updateaccount API.

        Args:
            action: Action to execute.
            data: Arguments for the action.

        Returns:
            dict: Action response.

        Raises:
            :py:class:`coto.clients.decoupled_account.ReauthException`: You have to
                reauthenticate, then try again.
        """
        if not data:
            data = {}

        data['action'] = action
        data['redirect_uri'] = self._REDIRECT_URL
        data['csrf'] = self._csrf_token()

        r = self.session()._post(
            'https://signin.aws.amazon.com/updateaccount',
            data=data,
        )

        if r.status_code != 200:
            print(r.text)
            raise Exception("failed action {0}".format(action))

        out = json.loads(r.text)
        if out['state'] == 'FAIL' and out['properties']['action'] == 'reAuth':
            raise ReauthException()

        if out['state'].lower() != 'success':
            if 'Message' in out['properties']:
                raise Exception("failed action {0}: {1}".format(action, out['properties']['Message']))
            else:
                raise Exception("failed action {0}".format(action))

        return out['properties']

    def get_account_info(self):
        """
        Gets the account name and email address.

        Request Syntax:
            .. code-block:: python

                response = client.get_auth_state()

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'accountEmail': str,
                    'Message': str,
                    'accountName': str,
                    'Title': str,
                }

        Raises:
            :py:class:`coto.clients.decoupled_account.ReauthException`: You have to
                reauthenticate, then try again.
        """
        return self._action('getAuthState')

    def update_account_name(self, AccountName):
        """
        Sets a new account name.

        Request Syntax:
            .. code-block:: python

                response = client.update_account_name(
                    AccountName=str,
                )

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'Message': str,
                    'Title': str,
                    'updatedAccountName': str
                }

        Raises:
            :py:class:`coto.clients.decoupled_account.ReauthException`: You have to
                reauthenticate, then try again.
        """
        return self._action('updateAccountName', {
            'newAccountName': AccountName,
        })

    def update_account_email(self, Password, AccountEmail):
        """
        Sets a new account email address.

        Request Syntax:
            .. code-block:: python

                response = client.update_account_email(
                    Password=str,
                    AccountEmail=str,
                )

        Args:
            Password (str): Current account password.
            AccountEmail (str): New account email.

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'Message': str,
                    'Title': str,
                    'updatedAccountEmail': str
                }

        Raises:
            :py:class:`coto.clients.decoupled_account.ReauthException`: You have to
                reauthenticate, then try again.
        """
        return self._action('updateAccountEmail', {
            'password': Password,
            'newEmailAddress': AccountEmail,
        })

    def update_account_password(self, Password, NewPassword):
        """
        Sets a new account password.

        Request Syntax:
            .. code-block:: python

                response = client.update_account_password(
                    Password=str,
                    NewPassword=str,
                )

        Args:
            Password (str): Current account password.
            NewPassword (str): New account password.

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'Message': str,
                    'Title': str,
                }

        Raises:
            :py:class:`coto.clients.decoupled_account.ReauthException`: You have to
                reauthenticate, then try again.
        """
        return self._action(
            'updateAccountPassword', {
                'oldpassword': Password,
                'newpassword': NewPassword,
                'newpassword1': NewPassword,
            })



import requests
import json
from urllib.parse import unquote
from colors import color
from .. import clients


def dr(r):
    for i in r.history + [r]:
        if i.status_code < 400:
            fg = 'green'
        else:
            fg = 'red'

        print(
            color(
                str(i.status_code) + \
                    " " + i.request.method + \
                    " " + i.request.url,
                fg=fg,
                style='underline'))

        for k, v in i.request.headers.items():
            if k == 'Cookie':
                print(color(k + ':', fg='blue'))
                for c in v.split(";"):
                    c = c.strip()
                    (n, c) = c.split('=', maxsplit=1)
                    print(color('    ' + n + ': ', fg='blue') + unquote(c))
            else:
                print(color(k + ':', fg='blue'), v)

        for k, v in i.headers.items():
            print(color(k + ':', fg='yellow'), v)

        if i.request.body and len(i.request.body) > 0:
            print(color('Body:', fg='blue'))
            print(i.request.body)
            print(color('EOF', fg='blue'))


class Session:
    """
    The Session class represents a session with the AWS Management Console.

    Use the `client` method to obtain a client for one of the supported
    services.
    """

    def __init__(
        self, debug=False, verify=True,
        metadata1_generator=None,
        captcha_solver=None, **kwargs
    ):
        """
        Args:
            debug (bool): Enable debug messages.
            verify (str | bool): Requests SSL certificate checking. Path to
                CA certificates file. ``False`` to ignore certificate errors.
                ``True`` to use defaults (default).
            captcha_solver (coto.captcha.Solver): Class implementing a way to solve captchas (e.g., send them to Slack for you to solve).
            metadata1_generator (coto.metadata1.Generator): Class implementing a way to generate metadata1.
            **kwargs: You can pass arguments for the signin method here.
        """
        self.debug = debug
        self._metadata1_generator = metadata1_generator
        self._captcha_solver = captcha_solver
        self.root = False
        self.coupled = None
        self.session = requests.Session()
        self.session.verify = verify
        self.authenticated = False
        self._clients = {}

        self.timeout = (3.1, 10)
        self.user_agent = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/64.0.3282.186 Safari/537.36'

        if len(kwargs) > 0:
            self.signin(**kwargs)

    def signin(self, **kwargs):
        """
        Signin to the AWS Management Console.

        There are various ways to sign in:
            * Using a boto3.Session object, pass the ``boto3_session`` argument.
            * Using the Account Root User, pass the ``email``, ``password``, and
              optionally ``mfa_secret`` arguments.

        Args:
            boto3_session (boto3.session.Session): The credentials of this
                session are retrieved and used to signin to the console.
            email (str): AWS account root user email to use for login.
            password (str): AWS account root user password to use for login.
            mfa_secret (str): AWS account root user mfa secret to use for login.
                The Base32 seed defined as specified in RFC3548.
                The Base32StringSeed is Base64-encoded.
        """
        if 'boto3_session' in kwargs:
            boto3_session = kwargs.get('boto3_session')
            return self.client('federation').signin(boto3_session)

        elif 'email' in kwargs and 'password' in kwargs:
            args = {}
            for key in ['email', 'password', 'mfa_secret']:
                if key in kwargs:
                    args[key] = kwargs.get(key)
            return self.client('signin').signin(**args)

    # http requests
    def _set_defaults(self, kwargs):
        if not 'timeout' in kwargs:
            kwargs['timeout'] = self.timeout

        if not 'headers' in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['User-Agent'] = self.user_agent

    def _get(self, url, **kwargs):
        self._set_defaults(kwargs)
        r = self.session.get(url, **kwargs)
        if self.debug:
            dr(r)
        return r

    def _post(self, url, **kwargs):
        self._set_defaults(kwargs)

        r = self.session.post(url, **kwargs)
        if self.debug:
            dr(r)
        return r

    def _put(self, url, **kwargs):
        self._set_defaults(kwargs)
        r = self.session.put(url, **kwargs)
        if self.debug:
            dr(r)
        return r

    def _delete(self, url, **kwargs):
        self._set_defaults(kwargs)
        r = self.session.delete(url, **kwargs)
        if self.debug:
            dr(r)
        return r

    def client(self, service):
        """
        Create a client for a service.

        Supported services:
          * ``account``
          * ``billing``
          * ``federation``
          * ``iam``
          * ``mfa``
          * ``resetpassword``
          * ``signin``
          * ``signin_amazon``
          * ``signin_aws``

        Args:
            service: name of the service, eg., `billing`

        Returns:
            object: service client
        """
        service = service.lower()

        if service not in self._clients:
            if not hasattr(clients, service):
                raise Exception("service {0} unsupported".format(service))

            klass = getattr(clients, service).Client

            if klass.REQUIRES_AUTHENTICATION and not self.authenticated:
                raise Exception(
                    "signin before creating {0} service client".format(
                        service))

            self._clients[service] = klass(self)

        return self._clients[service]



from bs4 import BeautifulSoup
from pyotp import TOTP
from urllib import parse
import json
from .. import BaseClient
from . import exceptions
import time


def captcha_decorator(func):
    def wrapper(*args, **kwargs):
        self = args[0]

        captcha_guess = kwargs.get('captcha_guess')
        solver = self.session()._captcha_solver
        guess_uuid = None

        while True:
            try:
                _kwargs = {k: v for k, v in kwargs.items()}
                if captcha_guess:
                    _kwargs['captcha_guess'] = captcha_guess
                return func(*args, **_kwargs)
            except exceptions.CaptchaRequiredException as e:
                if solver is None:
                    raise

                if guess_uuid and captcha_guess and captcha_guess.action == e.action:
                    solver.incorrect(guess_uuid)

                guess_uuid = solver.solve(url=e.CaptchaURL)

                while True:
                    guess = solver.result(guess_uuid)

                    if guess is None:
                        time.sleep(5)
                    else:
                        break
                captcha_guess = e.guess(guess)
                continue
            break

    return wrapper


class Client(BaseClient):
    REQUIRES_AUTHENTICATION = False
    _REDIRECT_URL = "https://console.aws.amazon.com/console/home?state=hashArgs%23&isauthcode=true"

    def __init__(self, session):
        super().__init__(session)
        self.__csrf_token = None
        self.__session_id = None

    def _csrf_token(self):
        if self.__csrf_token == None:
            self._get_tokens()

        return self.__csrf_token

    def _session_id(self):
        if self.__session_id == None:
            self._get_tokens()

        return self.__session_id

    def _get_tokens(self):
        r = self.session()._get(
            "https://signin.aws.amazon.com/signin?redirect_uri=https%3A%2F%2Fconsole.aws.amazon.com%2Fconsole%2Fhome%3Fstate%3DhashArgs%2523%26isauthcode%3Dtrue&client_id=arn%3Aaws%3Aiam%3A%3A015428540659%3Auser%2Fhomepage&forceMobileApp=0"
        )

        if r.status_code != 200:
            raise Exception("failed get tokens")

        soup = BeautifulSoup(r.text, 'html.parser')
        meta = {
            m['name']: m['content']
            for m in soup.find_all('meta') if 'name' in m.attrs
        }
        self.__csrf_token = meta['csrf_token']
        self.__session_id = meta['session_id']

    def _action(self, action, data=None, api="signin", captcha_guess=None):
        """
        Execute an action on the signin API.

        Args:
            action: Action to execute.
            data: Arguments for the action.

        Returns:
            dict: Action response.
        """
        if not data:
            data = {}

        data['action'] = action
        data['redirect_uri'] = self._REDIRECT_URL
        data['csrf'] = self._csrf_token()
        data['sessionId'] = self._session_id()

        if captcha_guess and captcha_guess.action == action:
            data['captcha_token'] = captcha_guess.captcha_token
            data['captchaObfuscationToken'] = \
                captcha_guess.captchaObfuscationToken
            data['captcha_guess'] = captcha_guess.guess

        r = self.session()._post(
            "https://signin.aws.amazon.com/{}".format(api),
            data=data,
        )

        if r.status_code != 200:
            raise Exception("failed action {}: {}".format(action, r.text))

        out = json.loads(r.text)
        state = out.get('state', 'none').lower()
        properties = out.get('properties', {})

        if properties.get('Captcha',
                          'false') == 'true' and action != "captcha":
            raise exceptions.CaptchaRequiredException(
                properties['CES'], properties['CaptchaURL'],
                properties['captchaObfuscationToken'], action)

        if state != 'success':
            if 'Message' in properties:
                raise Exception("failed action {}: {}".format(
                    action, properties['Message']))
            else:
                raise Exception("failed action {}: {}".format(action, r.text))

        return properties

    @captcha_decorator
    def get_account_type(self, email, captcha_guess=None):
        """
        Determine the type of account.

        Account Types:
            Coupled: Coupled to an amazon.com account.
            Decoupled: Independend from amazon.com.
            Unknown: Non-existent account.

        Request Syntax:

            .. code-block:: python

                response = client.get_account_type(
                    email=str,
                )

        Args:
            email: Account email address.

        Returns:
            str: Account type
        """
        response = self._action(
            'resolveAccountType', {'email': email},
            captcha_guess=captcha_guess)
        return response['resolvedAccountType']

    def mfa_required(self, email):
        mfa_client = self.session().client('mfa')
        mfa = mfa_client.get_mfa_status(email)
        if 'mfaType' in mfa:
            if mfa['mfaType'] == 'NONE':
                return False
            else:
                return True

            return True

        return True

    def signin(self, email, password, mfa_secret=None):
        # check mfa
        mfa_required = self.mfa_required(email)
        if mfa_required and (mfa_secret is None or len(mfa_secret) == 0):
            raise Exception("account mfa protected but no secret provided")

        if not mfa_required:
            mfa_secret = None

        return self.signin_decoupled(email, password, mfa_secret)

    @captcha_decorator
    def signin_decoupled(self,
                         email,
                         password,
                         mfa_secret=None,
                         captcha_guess=None):
        """
        Signin into the AWS Management Console using account root user.

        Request Syntax:

            .. code-block:: python

                response = client.signin_decoupled(
                    email=str,
                    password=str,
                    mfa_secret=str,
                )

        Args:
            email: Account email address.
            password: Account password.
            mfa_secret: Account mfa secret. The Base32 seed defined as specified
                in RFC3548. The Base32StringSeed is Base64-encoded.

        Returns:
            bool: Signin successful
        """
        data = {
            'email': email,
            'password': password,
            'client_id': 'arn:aws:iam::015428540659:user/homepage',
        }

        if mfa_secret is not None:
            data['mfaType'] = 'OTP'
            data['mfa1'] = TOTP(mfa_secret).now()
            data['mfaSerial'] = 'undefined'

        # an exception is thrown if authentication was unsuccessful
        self._action('authenticateRoot', data, captcha_guess=captcha_guess)
        self.session().authenticated = True
        self.session().root = True
        return True

    def get_password_recovery_captcha(self):
        """
        Obtains a captcha for password recovery.

        The value ``CES`` must be used as ``captcha_token`` in
        :py:meth:`get_reset_password_token`.

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'CES': str,
                    'Captcha': bool,
                    'CaptchaURL': str,
                    'captchaObfuscationToken': str,
                }
        """
        return self._action('captcha', {'forgotpassword': True})

    def raise_password_recovery_captcha(self):
        """
        Obtains a captcha for password recovery and raises a
        CaptchaRequiredException.
        """
        captcha = self.get_password_recovery_captcha()
        raise exceptions.CaptchaRequiredException(
            captcha['CES'], captcha['CaptchaURL'],
            captcha['captchaObfuscationToken'], 'getResetPasswordToken')

    @captcha_decorator
    def get_reset_password_token(self, email, captcha_guess=None):
        """
        Asks for a password reset token to be sent to the registered email
        address.

        The value token url from the resulting email must be used as
        ``reset_token_url`` in :py:meth:`reset_password`.

        Returns:
            dict: Response Syntax

            .. code-block:: python

                {
                    'recovery_result': 'email_sent'
                }
        """

        if not captcha_guess:
            self.raise_password_recovery_captcha()

        try:
            return self._action(
                'getResetPasswordToken', {'email': email},
                captcha_guess=captcha_guess)
        except Exception as e:
            if str(
                    e
            ) == "failed action getResetPasswordToken: Enter the characters and try again":
                self.raise_password_recovery_captcha()



from setuptools import setup, find_packages

setup(
    name = "coto",
    url = "https://github.com/sentialabs/coto",
    description = "Undocumented AWS Mananagement Console API Client",
    long_description_content_type='text/x-rst',
    long_description = """
        Some things do not have a documented API but can be set using the
        AWS Mananagement Console. Using this client we access the undocumented
        REST APIs that power the AWS Management Console.
    """,
    author = "Sentia MPC B.V.",
    author_email = "info@sentia.com",
    license = "Apache",
    version = "0.3.0",
    packages = find_packages(),
    install_requires = [
        'requests',
        'furl',
        'ansicolors',
        'beautifulsoup4',
        'pyotp',
    ],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords="aws boto",
    project_urls={
        'Documentation': 'http://coto.readthedocs.io/',
        'Source': 'https://github.com/sentialabs/coto',
    },
)
